package ru.dbotthepony.mc.otm.item

import com.google.common.collect.ImmutableMultimap
import com.google.common.collect.Multimap
import net.minecraft.ChatFormatting
import net.minecraft.core.BlockPos
import net.minecraft.core.NonNullList
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.entity.EquipmentSlot
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.ai.attributes.Attribute
import net.minecraft.world.entity.ai.attributes.AttributeModifier
import net.minecraft.world.entity.ai.attributes.Attributes
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.CreativeModeTab
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraftforge.common.ForgeConfigSpec
import net.minecraftforge.common.ToolAction
import net.minecraftforge.common.ToolActions
import net.minecraftforge.common.capabilities.ICapabilityProvider
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.capability.EnergyConsumerItem
import ru.dbotthepony.mc.otm.capability.ItemEnergyStorageImpl
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.extractEnergyInnerExact
import ru.dbotthepony.mc.otm.capability.getBarColor
import ru.dbotthepony.mc.otm.capability.getBarWidth
import ru.dbotthepony.mc.otm.capability.matteryEnergy
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.ImpreciseFractionConfigValue
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.defineImpreciseFraction
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.registry.EMPDamageSource
import ru.dbotthepony.mc.otm.core.WriteOnce

/**
 * This is called from [net.minecraft.world.item.enchantment.EnchantmentHelper.getSweepingDamageRatio]
 * by coremod patch
 */
fun getSweepingDamageRatioHook(ply: LivingEntity): Float? {
	if (ply.mainHandItem.item is EnergySwordItem && ply.mainHandItem.matteryEnergy?.extractEnergyInnerExact(EnergySwordItem.ENERGY_PER_SWING, true)?.isPositive == true) {
		return 1f
	}

	return null
}

class EnergySwordItem : Item(Properties().stacksTo(1).rarity(Rarity.RARE).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
	val chargedAttributes: Multimap<Attribute, AttributeModifier>
	val dischargedAttributes: Multimap<Attribute, AttributeModifier>

	init {
		var builder = ImmutableMultimap.builder<Attribute, AttributeModifier>()

		builder.put(Attributes.ATTACK_DAMAGE, AttributeModifier(BASE_ATTACK_DAMAGE_UUID, "Weapon modifier", 11.0, AttributeModifier.Operation.ADDITION))
		builder.put(Attributes.ATTACK_SPEED, AttributeModifier(BASE_ATTACK_SPEED_UUID, "Weapon modifier", -2.4, AttributeModifier.Operation.ADDITION))

		chargedAttributes = builder.build()

		builder = ImmutableMultimap.builder()

		builder.put(Attributes.ATTACK_DAMAGE, AttributeModifier(BASE_ATTACK_DAMAGE_UUID, "Weapon modifier", 3.5, AttributeModifier.Operation.ADDITION))
		builder.put(Attributes.ATTACK_SPEED, AttributeModifier(BASE_ATTACK_SPEED_UUID, "Weapon modifier", -2.4, AttributeModifier.Operation.ADDITION))

		dischargedAttributes = builder.build()
	}

	override fun getDestroySpeed(itemStack: ItemStack, blockState: BlockState): Float {
		val energy = itemStack.getCapability(MatteryCapability.ENERGY).orNull() ?: return 1f

		if (blockState.`is`(Blocks.COBWEB)) {
			if (energy.batteryLevel < COBWEB_POWER_COST) {
				return 2f
			} else {
				return 25f
			}
		}

		return when (blockState.material) {
			Material.PLANT, Material.REPLACEABLE_PLANT, Material.VEGETABLE -> if (energy.batteryLevel < PLANT_POWER_COST) 1.5f else 4f
			Material.LEAVES -> if (energy.batteryLevel < PLANT_POWER_COST) 1.5f else 8f
			else -> 1f
		}
	}

	override fun canAttackBlock(p_41441_: BlockState, p_41442_: Level, p_41443_: BlockPos, p_41444_: Player): Boolean {
		return !p_41444_.isCreative
	}

	override fun hurtEnemy(itemStack: ItemStack, victim: LivingEntity, attacker: LivingEntity): Boolean {
		if (attacker is Player && attacker.isCreative) {
			victim.matteryPlayer?.let {
				if (it.isAndroid) {
					victim.hurt(EMPDamageSource(attacker), 8f)
				}
			}

			return true
		}

		itemStack.getCapability(MatteryCapability.ENERGY).ifPresentK {
			if (!it.extractEnergyInnerExact(ENERGY_PER_SWING, false).isZero) {
				victim.matteryPlayer?.let {
					if (it.isAndroid) {
						it.androidEnergy.extractEnergyInner(ENERGY_ZAP, false)
						victim.hurt(EMPDamageSource(attacker), 8f)
					}
				}
			}
		}

		return true
	}

	override fun isBarVisible(p_150899_: ItemStack): Boolean {
		return p_150899_.matteryEnergy != null
	}

	override fun getBarWidth(p_150900_: ItemStack): Int {
		return p_150900_.matteryEnergy?.getBarWidth() ?: super.getBarWidth(p_150900_)
	}

	override fun getBarColor(p_150901_: ItemStack): Int {
		return p_150901_.matteryEnergy?.getBarColor() ?: super.getBarColor(p_150901_)
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		p_41423_: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, p_41423_, p_41424_)

		ItemEnergyStorageImpl.appendHoverText(itemStack, p_41423_)

		p_41423_.add(DESCRIPTION)
		p_41423_.add(DESCRIPTION2)
		p_41423_.add(DESCRIPTION3)
		p_41423_.add(DESCRIPTION4)
	}

	override fun mineBlock(
		itemStack: ItemStack,
		p_41417_: Level,
		blockState: BlockState,
		p_41419_: BlockPos,
		p_41420_: LivingEntity
	): Boolean {
		if (blockState.getDestroySpeed(p_41417_, p_41419_) != 0f && (p_41420_ !is Player || !p_41420_.isCreative)) {
			val energy = itemStack.matteryEnergy

			when (blockState.material) {
				Material.PLANT, Material.REPLACEABLE_PLANT, Material.VEGETABLE, Material.LEAVES ->
					energy?.extractEnergyInnerExact(PLANT_POWER_COST, false)
			}

			if (blockState.`is`(Blocks.COBWEB)) {
				energy?.extractEnergyInnerExact(COBWEB_POWER_COST, false)
			}
		}

		return true
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return EnergyConsumerItem(stack, MAX_ENERGY)
	}

	override fun fillItemCategory(p_41391_: CreativeModeTab, p_41392_: NonNullList<ItemStack>) {
		super.fillItemCategory(p_41391_, p_41392_)

		if (allowedIn(p_41391_)) {
			p_41392_.add(ItemStack(this).also {
				it.matteryEnergy?.receiveEnergyInner(MAX_ENERGY, false)
			})
		}
	}

	override fun getAttributeModifiers(
		slot: EquipmentSlot,
		itemStack: ItemStack
	): Multimap<Attribute, AttributeModifier> {
		if (slot != EquipmentSlot.MAINHAND) {
			return ImmutableMultimap.of()
		}

		val energy = itemStack.matteryEnergy

		if (energy != null && energy.batteryLevel >= ENERGY_PER_SWING) {
			return chargedAttributes
		}

		return dischargedAttributes
	}

	override fun canPerformAction(stack: ItemStack, toolAction: ToolAction): Boolean {
		if (stack.matteryEnergy?.extractEnergyInnerExact(ENERGY_PER_SWING, true)?.isPositive == true) {
			return ToolActions.DEFAULT_SWORD_ACTIONS.contains(toolAction)
		}

		return false
	}

	companion object {
		val MAX_ENERGY get() = _MAX_ENERGY.get()
		val ENERGY_ZAP get() = _ENERGY_ZAP.get()
		val ENERGY_PER_SWING get() = _ENERGY_PER_SWING.get()
		val COBWEB_POWER_COST get() = _COBWEB_POWER_COST.get()
		val PLANT_POWER_COST get() = _PLANT_POWER_COST.get()

		private var _MAX_ENERGY: ImpreciseFractionConfigValue by WriteOnce()
		private var _ENERGY_ZAP: ImpreciseFractionConfigValue by WriteOnce()
		private var _ENERGY_PER_SWING: ImpreciseFractionConfigValue by WriteOnce()
		private var _COBWEB_POWER_COST: ImpreciseFractionConfigValue by WriteOnce()
		private var _PLANT_POWER_COST: ImpreciseFractionConfigValue by WriteOnce()

		fun registerConfig(builder: ForgeConfigSpec.Builder) {
			builder.comment("Energy sword values").push("energy_sword")

			_MAX_ENERGY = builder.defineImpreciseFraction("MAX_ENERGY", ImpreciseFraction(500_000), ImpreciseFraction.ZERO)
			_ENERGY_ZAP = builder.defineImpreciseFraction("ENERGY_ZAP", ImpreciseFraction(4_000), ImpreciseFraction.ZERO)
			_ENERGY_PER_SWING = builder.defineImpreciseFraction("ENERGY_PER_SWING", ImpreciseFraction(2_000), ImpreciseFraction.ZERO)
			_COBWEB_POWER_COST = builder.defineImpreciseFraction("COBWEB_POWER_COST", ImpreciseFraction(2_500), ImpreciseFraction.ZERO)
			_PLANT_POWER_COST = builder.defineImpreciseFraction("PLANT_POWER_COST", ImpreciseFraction(500), ImpreciseFraction.ZERO)

			builder.pop()
		}

		private val DESCRIPTION = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc").withStyle(ChatFormatting.DARK_GRAY)
		private val DESCRIPTION2 = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc2").withStyle(ChatFormatting.DARK_GRAY)
		private val DESCRIPTION3 = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc3").withStyle(ChatFormatting.DARK_GRAY)
		private val DESCRIPTION4 = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc4").withStyle(ChatFormatting.DARK_GRAY)
	}
}
